(* $Id: gont.ml,v 1.10 2001/11/12 09:09:25 malekith Exp $ *)

type location = {mutable file: string; mutable line: int}

(* Identifier. *)
type id = string

(* Types used by gont compiler. *)
type ty =
    Int                         (* integer of the same size as pointer *)
  | Void                        (* unit type *)
  | Bool                        (* boolean, same size as Int *)
  | Float                       (* C's double *)
  | Func of ty * (ty list)
  | RealFunc of ty * (ty list)
  | String

type target_int = int
type target_float = float

type raw_exp =
    SimpleBinOp of Ksi.binop * exp * exp
  | SimpleUnOp of Ksi.unop * exp
  | Var of id
  | Call of exp * exp list
  | IntConst of string
  | BoolConst of bool
  | FloatConst of string
  | StringConst of string
and exp = {mutable raw_exp: raw_exp; 
           mutable exp_ty: ty option; 
           exp_loc: location}
and raw_stmt =
    Skip
  | Exp of exp
  | Return of exp
  | ReturnVoid
  | Block of stmt list
and stmt = {raw_stmt: raw_stmt;
            stmt_loc: location}

type storage =
    Local
  | Public

type fn_decl = {fn_storage: storage;
                fn_ret_type: ty; 
                fn_name: id; 
                fn_parms: (ty * id) list;
                fn_body: stmt;
                fn_loc: location}
type extern_c_decl = {xc_ret_type: ty; 
                      xc_name: id; 
                      xc_parms: ty list;
                      xc_loc: location}
type decl =
    FuncDecl of fn_decl
  | GlobalDecl of storage * ty * id * location
  | ExternC of extern_c_decl

exception Syntax_error of location * string
exception Eof
;;

let xf = Printf.sprintf

let err_msg (loc, msg) =
  prerr_string (xf "%s:%d: %s" loc.file loc.line msg);
  prerr_newline ()

let warn (loc, msg) = err_msg (loc, ("warning: " ^ msg))

let err (loc, msg) = raise (Syntax_error(loc, msg))
;;

let rec ksi_type t =
  match t with
      Int -> Ksi.ptroff ()
    | Void -> Ksi.Void
    | Bool -> Ksi.ptroff ()     (* bool is C's int *)
    | Float -> Ksi.c_type(Ksi.C_double)
    | RealFunc(r, a) -> Ksi.FuncPtr(ksi_type r, (List.map ksi_type a))
    | Func(_, _) -> Ksi.Ptr(Ksi.Void)
    | String -> Ksi.Ptr(Ksi.UInt(8))
;;

let debug s = print_string s; print_newline ();;

class symbol = fun (t, n) ->
  object
    val sym_type : ty = t
    val sym_name : id = n
    method get_name = sym_name
    method get_type = sym_type
    (* Probably we'll need other info here as well.
     * That's why symbol table ain't simply (id, ty) Hashtbl.t 
     *)
  end

class symbol_table = 
  object
    val mutable hash : (id, symbol) Hashtbl.t = (Hashtbl.create 16)
    method get_symbol n = Hashtbl.find hash n
    method add_symbol s = Hashtbl.add hash s#get_name s
    method copy_self = 
      let c = new symbol_table in
        Hashtbl.iter (fun _ s -> c#add_symbol (Oo.copy s)) hash; c
  end

exception SymResult of symbol

class compenv =
  object (self)
    val syms : symbol_table Stack.t = 
      let s = Stack.create () in
        (Stack.push (new symbol_table) s; s)
    method lookup_symbol n = 
      try
        Stack.iter 
          (fun s ->
            try
              raise (SymResult (s#get_symbol n))
            with Not_found -> ()) 
          syms;
        raise Not_found
      with SymResult(r) -> r
    method symbol_exists n =
      try
        ignore(self#lookup_symbol n);
        true
      with Not_found -> false
    method symbol_exists_in_current_scope n = 
      try
        ignore((Stack.top syms)#get_symbol n);
        true
      with Not_found -> false
    method add_symbol s = (Stack.top syms)#add_symbol s
    method push_level = Stack.push (new symbol_table) syms
    method pop_level = ignore(Stack.pop syms)

    val decls : Ksi.decl Stack.t = Stack.create ()
    method add_decl d = Stack.push d decls
    method get_decls = 
      let rec stack2list xs =
        if (Stack.length decls) == 0 then xs
        else stack2list ((Stack.pop decls)::xs) in
      stack2list []
    
    val cnt = ref 0
    method seq_name s = incr cnt; ("@" ^ s ^ string_of_int(!cnt))

    val mutable data_fn_name = ""
    val data_fn_ret_type = ref Void
    method set_fn_name n = data_fn_name <- n
    method fn_name = data_fn_name
    method set_fn_ret_type t = data_fn_ret_type := t
    method fn_ret_type = !data_fn_ret_type

    (* Clear up... *)
    method fn_finish = data_fn_name <- ""
  end

let env = new compenv
;;


(* Check whatever t is subtype of s (t < s). At this very moment this 
 * checks for type equality... *)
let rec list_subtype tl sl =
  match tl with
      (t::ts) -> 
        (match sl with
            (s::ss) -> (subtype t s) && (list_subtype ts ss)
          | [] -> false)
    | [] ->
        (match sl with
            (_::_) -> false
           | [] -> true)
and subtype t s =
  match s with
      Void -> t = s
    | Int -> t = s
    | Float -> t = s
    | Bool -> t = s
    | String -> t = s
    | Func(t_s, a_s) -> 
        (match t with
            Func(t_t, a_t) -> (subtype t_t t_s) && (list_subtype a_t a_s)
          | _ -> false)
    | RealFunc(_, _) -> assert false
;;

let rec string_of_type_list tt =
  match tt with
      [t] -> string_of_type t
    | t::ts -> (string_of_type t) ^ ", " ^ (string_of_type_list ts)
    | [] -> ""
and string_of_type t =
  match t with
      Void -> "void"
    | Int -> "int"
    | Float -> "float"
    | Bool -> "bool"
    | String -> "string"
    | Func(t, a) -> 
        "[" ^ (string_of_type_list a) ^ "] -> " ^ (string_of_type t)
    | RealFunc(_, _) -> assert false
;;

let call_type_error actual formal =
  let rec typecheck_call no actuall formall =
    match formall with
        (formal::formals) ->
          (match actuall with
              (actual::actuals) -> 
                if subtype actual formal then 
                  (typecheck_call (no + 1) formals actuals)
                else
                  Some (xf "%s is not subtype of %s in parameter %d of call"
                           (string_of_type actual)
                           (string_of_type formal)
                           no)
            | _ -> assert false)
      | [] -> None in
  if (List.length formal) != (List.length actual) then
    Some (xf "there is %d formal parameters and %d actual parameters in call" 
             (List.length formal) 
             (List.length actual))
  else
    typecheck_call 1 actual formal
;;

(* This is going to prepend module name, mangle name or whatever. *)
let name_convert s = s
;;

let rec same_arith_types loc name e1 e2 =
  let (t1, t2) = (type_of e1, type_of e2) in
  if (subtype t1 t2) && (subtype t2 t1) then
    match t1 with
        Int -> Int
      | Float -> Float
      | _ -> err (loc, xf "type %s isn't arithmetical (for binary %s)"
                          (string_of_type t1)
                          name)
  else
    err (loc, xf "binary %s expects the same arithmetical \ 
                  type on both arguments, while types %s and %s \
                  are incompatibile"
                  name
                  (string_of_type t1)
                  (string_of_type t2))
and same_int_types loc name e1 e2 =
  let (t1, t2) = (type_of e1, type_of e2) in
  if (subtype t1 t2) && (subtype t2 t1) then
    match t1 with
        Int -> Int
      | _ -> err (loc, xf "type %s isn't integer (for binary %s)"
                          (string_of_type t1)
                          name)
  else
    err (loc, xf "binary %s expects the same integer \
                  type on both arguments, while types %s and %s \
                  are incompatibile"
                  name
                  (string_of_type t1)
                  (string_of_type t2))
and typecheck_simple_unop loc (op, e) =
  match op with
      Ksi.Negate -> 
        (match (type_of e) with
            Int -> Int
          | Float -> Float
          | x -> err (loc, xf "unary - applaied to non-arithmetic type %s"
                              (string_of_type x)))
    | _ -> assert false
    
and typecheck_simple_binop e (op, e1, e2) =
  let oops s = err (e.exp_loc, s) in
  match op with
      Ksi.Mult -> same_arith_types e.exp_loc "*" e1 e2
    | Ksi.TruncDiv -> 
      let t = same_arith_types e.exp_loc "/" e1 e2 in
      if subtype t Float then
        (e.raw_exp <- SimpleBinOp(Ksi.RDiv, e1, e2); t)
      else
        t
    | Ksi.TruncMod -> same_int_types e.exp_loc "%" e1 e2
    | Ksi.Plus -> same_arith_types e.exp_loc "+" e1 e2
    | Ksi.Minus -> same_arith_types e.exp_loc "-" e1 e2
    | _ -> assert false
    
and typecheck e =
  let oops s = err (e.exp_loc, s) in
  match e.raw_exp with
    SimpleBinOp(op, e1, e2) ->
      typecheck_simple_binop e (op, e1, e2)
  | SimpleUnOp(op, e) ->
      typecheck_simple_unop e.exp_loc (op, e)
  | Var n ->
      (try
         match (env#lookup_symbol n)#get_type with
             RealFunc(r, a) -> Func(r, a)
           | x -> x
       with Not_found ->
         oops (xf "use of undefined symbol `%s'" n))
  | Call(f, act) ->
      (match type_of f with
          Func(r, form) -> 
             (match call_type_error (List.map type_of act) form with
                 None -> r
               | Some err -> oops err)
        | _ -> oops "called object is not a function")
  | IntConst _ -> Int
  | BoolConst _ -> Bool
  | FloatConst _ -> Float
  | StringConst _ -> String
  
and type_of e =
  match e.exp_ty with
      Some t -> t
    | None -> e.exp_ty <- Some (typecheck e); type_of e
;;

let rec gen_exp e =
  (* This simple line does most of the work. It ensures
   * whatever given operator can take parameters of this type,
   * whatever all variables are defined, and so on. *)
  ignore(type_of e);
  (* Now the easy part. Map to Ksi code. *)
  match e.raw_exp with
      SimpleBinOp(op, e1, e2) -> Ksi.Binary(op, gen_exp e1, gen_exp e2)
    | SimpleUnOp(op, e) -> Ksi.Unary(op, gen_exp e)
    | Var(n) -> 
        (match (env#lookup_symbol n)#get_type with
             RealFunc(_, _) -> Ksi.FuncRef(n)
           | _ -> Ksi.Ref(n))
    | Call(e, p) -> Ksi.PtrCall(gen_exp e, (List.map gen_exp p))
    | IntConst(i) -> Ksi.IntConst(ksi_type Int, i)
    | BoolConst(x) -> Ksi.BoolConst(x)
    | FloatConst(f) -> Ksi.FloatConst(ksi_type Float, f)
    | StringConst(s) -> Ksi.StringConst(s)
  
and gen_parms loc pp =
  match pp with
      (t, n)::ps -> 
        if env#symbol_exists_in_current_scope n then
          err (loc, xf "parameter `%s' given twice" n)
        else if env#symbol_exists n then
          warn (loc, xf "parameter `%s' shadows other symbol" n);
        env#add_symbol (new symbol(t, n));
        (ksi_type t, n) :: (gen_parms loc ps)
    | [] -> []

and gen_body s =
  match s.raw_stmt with
      Skip -> Ksi.Nop
    | Exp(e) -> gen_exp e
    | Return(e) -> 
      if subtype (type_of e) (env#fn_ret_type) then
        Ksi.Return(gen_exp e)
      else
        err (s.stmt_loc, xf "returned value is of type %s, \
                             while it should be %s"
                            (string_of_type (type_of e))
                            (string_of_type env#fn_ret_type))
    | ReturnVoid ->
      if subtype Void (env#fn_ret_type) then
        Ksi.Return(Ksi.Nop)
      else
        err (s.stmt_loc, "this function should return a value")
    | Block([]) -> Ksi.Nop
    | Block([s]) -> Ksi.Bind(gen_body s)
    | Block([s1; s2]) -> Ksi.Bind(Ksi.Compound(gen_body s1, gen_body s2))
    | Block(s) -> Ksi.Bind(Ksi.CompoundList(List.map gen_body s))

and codegen d =
  let code = 
    let convert_storage s =
      match s with
          Local -> [Ksi.Local]
        | Public -> [Ksi.Public] in
    match d with
        FuncDecl(f) -> 
           env#push_level;
           env#set_fn_ret_type f.fn_ret_type;
           env#set_fn_name f.fn_name;
           (let fntype = RealFunc(env#fn_ret_type, 
                                 (List.map (fun (t, _) -> t) f.fn_parms)) in
            env#add_symbol (new symbol(fntype, env#fn_name)));
           let parms = gen_parms f.fn_loc f.fn_parms in ();
           let body = gen_body f.fn_body in ();
           env#pop_level;
           let c = Ksi.Func(convert_storage f.fn_storage,
                            ksi_type env#fn_ret_type,
                            (name_convert env#fn_name),
                            parms,
                            body) in
           (env#fn_finish; c)
      | GlobalDecl(s, t, n, _) ->
           env#add_symbol (new symbol(t, n));
           Ksi.Global(convert_storage s, 
                      ksi_type t, 
                      name_convert n)
      | ExternC(x) -> 
           env#add_symbol (new symbol(RealFunc(x.xc_ret_type, x.xc_parms),
                                      x.xc_name));
           Ksi.FuncProto([Ksi.Extern], 
                         ksi_type x.xc_ret_type,
                         x.xc_name,
                         (List.map ksi_type x.xc_parms)) in
  code :: env#get_decls
;;
